/*
 * Copyright (c) 1999 Greg Haerr <greg@censoft.com>
 *
 * Win32 API upper level graphics drawing routines
 */
#include "windows.h"
#include "wintern.h"
#include <stdlib.h>

#define MAXSYSCOLORS	29	/* # of COLOR_* system colors*/
#define MAXSTOCKOBJECTS	18	/* # of stock objects*/

static HDC	cliphdc;	/* current window cliprects*/

HDC WINAPI
GetDC(HWND hwnd)
{
	HDC	hdc;

	// add caching?
	hdc = ItemNew(struct hdc);
	if(!hdc)
		return NULL;

	if(!hwnd)		/* handle NULL hwnd => desktop*/
		hwnd = rootwp;

	hdc->hwnd = hwnd;
	hdc->bClientDC = TRUE;
	hdc->bkmode = OPAQUE;
	hdc->bkcolor = WHITE;
	hdc->textcolor = BLACK;
	hdc->brush = GetStockObject(WHITE_BRUSH);
	hdc->pen = GetStockObject(BLACK_PEN);
	hdc->font = GetStockObject(SYSTEM_FONT);
	hdc->pt.x = 0;
	hdc->pt.y = 0;

	hdc->mode = MODE_SET;

	return hdc;
}

HDC WINAPI
GetWindowDC(HWND hwnd)
{
	HDC	hdc;

	hdc = GetDC(hwnd);
	if(hdc)
		hdc->bClientDC = FALSE;
	return hdc;
}

int WINAPI 
ReleaseDC(HWND hwnd, HDC hdc)
{
	if(!hdc)
		return 0;
	if(hdc == cliphdc)
		cliphdc = NULL;
	DeleteObject(hdc->brush);
	DeleteObject(hdc->pen);
	DeleteObject(hdc->font);
	ItemFree(hdc);
	return 1;
}

HDC WINAPI 
BeginPaint(HWND hwnd, LPPAINTSTRUCT lpPaint)
{
	HDC	hdc;

	/* first update non-client area*/
	SendMessage(hwnd, WM_NCPAINT, 0, 0L);

	hdc = GetDC(hwnd);
	/* then erase client background*/
	lpPaint->fErase = SendMessage(hwnd, WM_ERASEBKGND, (WPARAM)hdc, 0L);
	lpPaint->hdc = hdc;

	/* update region is full client area for now*/
	GetClientRect(hwnd, &lpPaint->rcPaint);
	return hdc;
}

BOOL WINAPI 
EndPaint(HWND hwnd, CONST PAINTSTRUCT *lpPaint)
{
	ReleaseDC(hwnd, lpPaint->hdc);
	return TRUE;
}

COLORREF WINAPI
SetTextColor(HDC hdc, COLORREF crColor)
{
	COLORREF	oldtextcolor;

	if (!hdc)
		return crColor;
	oldtextcolor = hdc->textcolor;
	hdc->textcolor = (COLORVAL)crColor;
	return oldtextcolor;
}

COLORREF WINAPI
SetBkColor(HDC hdc, COLORREF crColor)
{
	COLORREF	oldbkcolor;

	if (!hdc)
		return crColor;
	oldbkcolor = hdc->bkcolor;
	hdc->bkcolor = crColor;
	return oldbkcolor;
}

int WINAPI
SetBkMode(HDC hdc, int iBkMode)
{
	int	oldbkmode;

	if(!hdc)
		return iBkMode;
	oldbkmode = hdc->bkmode;
	hdc->bkmode = iBkMode;
	return oldbkmode;
}

/* 
 * Setup clip region from device context's associated window
 * return the DC's hwnd if window is visible
 */
HWND
WndPrepareDC(HDC hdc)
{
	HWND		hwnd;

	if(!hdc || !hdc->hwnd)
		return NULL;

	hwnd = hdc->hwnd;
	if (hwnd->unmapcount)
		return NULL;

	/*
	 * If the window is not the currently clipped one, then
	 * make it the current one and define its clip rectangles.
	 */
	if(hdc != cliphdc) {
		WndSetClipWindow(hdc);
		cliphdc = hdc;
	}

	return hwnd;
}

COLORREF WINAPI
SetPixel(HDC hdc, int x, int y, COLORREF crColor)
{
	HWND		hwnd;
	POINT		pt;

	hwnd = WndPrepareDC(hdc);
	if(!hwnd)
		return 0;	/* doesn't return previous color*/
	pt.x = x;
	pt.y = y;
	if(WndIsClientDC(hdc))
		ClientToScreen(hwnd, &pt);

	/* draw point in passed color*/
	GdSetForeground(GdFindColor(crColor));
	GdPoint(pt.x, pt.y);
	return 0;		/* doesn't return previous color*/
}

BOOL WINAPI 
MoveToEx(HDC hdc, int x, int y, LPPOINT lpPoint)
{
	if(!hdc)
		return FALSE;
	if(lpPoint)
		*lpPoint = hdc->pt;
	hdc->pt.x = x;
	hdc->pt.y = y;
	return TRUE;
}

BOOL WINAPI
LineTo(HDC hdc, int x, int y)
{
	HWND		hwnd;
	POINT		beg, end;

	hwnd = WndPrepareDC(hdc);
	if(!hwnd)
		return FALSE;

	beg.x = hdc->pt.x;
	beg.y = hdc->pt.y;
	end.x = x;
	end.y = y;
	if(WndIsClientDC(hdc)) {
		ClientToScreen(hwnd, &beg);
		ClientToScreen(hwnd, &end);
	}

	/* draw line in current pen color*/
	if(hdc->pen->style != PS_NULL) {
		GdSetForeground(GdFindColor(hdc->pen->color));
#if 0
		fix = 0;
		if(beg.x == end.x)
			fix |= 1;
		if(beg.y == end.y)
			fix |= 2;
		if(fix&1)
			--end.y;
		if(fix&2)
			--end.x;
#endif
		// fix: don't draw last point
		GdLine(beg.x, beg.y, end.x, end.y);
	}
	hdc->pt.x = x;
	hdc->pt.y = y;
	return TRUE;
}

BOOL WINAPI
Rectangle(HDC hdc, int nLeft, int nTop, int nRight, int nBottom)
{
	HWND	hwnd;
	RECT	rc;

	hwnd = WndPrepareDC(hdc);
	if(!hwnd)
		return FALSE;

	SetRect(&rc, nLeft, nTop, nRight, nBottom);
	if(WndIsClientDC(hdc)) {
		ClientToScreen(hwnd, (LPPOINT)&rc.left);
		ClientToScreen(hwnd, (LPPOINT)&rc.right);
	}

	/* draw rectangle in current pen color*/
	if(hdc->pen->style != PS_NULL) {
		GdSetForeground(GdFindColor(hdc->pen->color));
		GdRect(rc.left, rc.top, rc.right - rc.left, rc.bottom - rc.top);
	}

	/* fill rectangle in current brush color*/
	if(hdc->brush->style != BS_NULL) {
		InflateRect(&rc, -1, -1);
		GdSetForeground(GdFindColor(hdc->brush->color));
		GdFillRect(rc.left, rc.top, rc.right - rc.left,
			rc.bottom - rc.top);
	}

	return TRUE;
}

int WINAPI
FillRect(HDC hdc, CONST RECT *lprc, HBRUSH hbr)
{
	HWND		hwnd;
	RECT 		rc;
	COLORREF	crFill;

	hwnd = WndPrepareDC(hdc);
	if(!hwnd || !hbr)
		return FALSE;

	if(!lprc) {
		if(WndIsClientDC(hdc))
			GetClientRect(hwnd, &rc);
		else
			GetWindowRect(hwnd, &rc);
		lprc = &rc;
	}
	if(WndIsClientDC(hdc)) {
		ClientToScreen(hwnd, (LPPOINT)lprc);
		ClientToScreen(hwnd, (LPPOINT)&lprc->right);
	}

	/* handle COLOR_xxx + 1 passed as HBRUSH*/
	if((int)hbr <= MAXSYSCOLORS)
		crFill = GetSysColor((int)hbr-1);
	else {
		/* get color from passed HBRUSH*/
		if(hbr->style == BS_NULL)
			return TRUE;
		crFill = hbr->color;
	}

	/* fill rectangle in passed brush color*/
	GdSetForeground(GdFindColor(crFill));
	GdFillRect(lprc->left, lprc->top, lprc->right - lprc->left,
		lprc->bottom - lprc->top);
	return TRUE;
}

BOOL WINAPI
TextOut(HDC hdc, int x, int y, LPCSTR lpszString, int cbString)
{
	return ExtTextOut(hdc, x, y, 0, NULL, lpszString, cbString, NULL);
}

BOOL WINAPI
ExtTextOut(HDC hdc, int x, int y, UINT fuOptions, CONST RECT *lprc,
	LPCSTR lpszString, UINT cbCount, CONST INT *lpDx)
{
	HWND	hwnd;
	POINT	pt;
	RECT	rc;

	hwnd = WndPrepareDC(hdc);
	if(!hwnd)
		return FALSE;

	pt.x = x;
	pt.y = y;
	if(WndIsClientDC(hdc))
		ClientToScreen(hwnd, &pt);

	/* optionally fill passed rectangle*/
	if(lprc && (fuOptions&ETO_OPAQUE)) {
		if(WndIsClientDC(hdc)) {
			rc = *lprc;
			ClientToScreen(hwnd, (LPPOINT)&rc.left);
			ClientToScreen(hwnd, (LPPOINT)&rc.right);

			/* fill rectangle with current background color*/
			GdSetForeground(hdc->bkcolor);
			GdFillRect(lprc->left, lprc->top,
				lprc->right - lprc->left,
				lprc->bottom - lprc->top);
		}
		GdSetUseBackground(FALSE);
	} else {
		/* use current background mode for text background draw*/
		GdSetUseBackground(hdc->bkmode == OPAQUE? TRUE: FALSE);
		if(hdc->bkmode == OPAQUE)
			GdSetBackground(GdFindColor(hdc->bkcolor));
	}

	/* nyi: lpDx*/

	/* draw text in current text foreground and background color*/
	GdSetForeground(GdFindColor(hdc->textcolor));
	GdSetFont(hdc->font->style);
	GdText(pt.x, pt.y, lpszString, cbCount, FALSE);

	return TRUE;
}

static COLORREF sysColors[MAXSYSCOLORS] = {
	RGB(224, 224, 224),	/* COLOR_SCROLLBAR           */
	RGB(128, 128, 128),     /* COLOR_BACKGROUND          */
	RGB(128,   0, 128),     /* COLOR_ACTIVECAPTION       */
	RGB(  0,  64, 128),  	/* COLOR_INACTIVECAPTION     */
	RGB(192, 192, 192),     /* COLOR_MENU                */
	RGB(255, 255, 255),     /* COLOR_WINDOW              */
	RGB(  0,   0,   0),     /* COLOR_WINDOWFRAME         */
	RGB(  0,   0,   0),     /* COLOR_MENUTEXT            */
	RGB(  0,   0,   0),     /* COLOR_WINDOWTEXT          */
	RGB(255, 255, 255),     /* COLOR_CAPTIONTEXT         */
	RGB(192, 192, 192),   	/* COLOR_ACTIVEBORDER        */
	RGB(192, 192, 192), 	/* COLOR_INACTIVEBORDER      */
	RGB(128, 128, 128),   	/* COLOR_APPWORKSPACE        */
	RGB(  0,   0, 128),     /* COLOR_HIGHLIGHT           */
	RGB(255, 255, 255),     /* COLOR_HIGHLIGHTTEXT       */
	RGB(192, 192, 192),     /* COLOR_BTNFACE             */
	RGB(128, 128, 128),   	/* COLOR_BTNSHADOW           */
	RGB(192, 192, 192),     /* COLOR_GRAYTEXT            */
	RGB(  0,   0,   0),     /* COLOR_BTNTEXT             */
	RGB(  0,   0,   0),    	/* COLOR_INACTIVECAPTIONTEXT */
	RGB(255, 255, 255),  	/* COLOR_BTNHIGHLIGHT        */
	RGB( 32,  32,  32),     /* COLOR_3DDKSHADOW          */
	RGB(192, 192, 192),     /* COLOR_3DLIGHT             */
	RGB(  0,   0,   0),     /* COLOR_INFOTEXT            */
	RGB(255, 255, 192), 	/* COLOR_INFOBK              */
	RGB(184, 180, 184),  	/* COLOR_ALTERNATEBTNFACE */
	RGB(  0,   0, 255),     /* COLOR_HOTLIGHT */
	RGB( 16, 132, 208),  	/* COLOR_GRADIENTACTIVECAPTION */
	RGB(184, 180, 184) 	/* COLOR_GRADIENTINACTIVECAPTION */
};

DWORD WINAPI
GetSysColor(int nIndex)
{
	if(nIndex >= 0 && nIndex < MAXSYSCOLORS)
		return sysColors[nIndex];
	return 0;
}

static struct hgdiobj OBJ_WHITE_BRUSH = {
	OBJ_BRUSH, BS_SOLID, RGB(255, 255, 255), TRUE
};

static struct hgdiobj OBJ_LTGRAY_BRUSH = {
	OBJ_BRUSH, BS_SOLID, RGB(192, 192, 192), TRUE
};

static struct hgdiobj OBJ_GRAY_BRUSH = {
	OBJ_BRUSH, BS_SOLID, RGB(128, 128, 128), TRUE
};

static struct hgdiobj OBJ_DKGRAY_BRUSH = {
	OBJ_BRUSH, BS_SOLID, RGB(32, 32, 32), TRUE
};

static struct hgdiobj OBJ_BLACK_BRUSH = {
	OBJ_BRUSH, BS_SOLID, RGB(0, 0, 0), TRUE
};

static struct hgdiobj OBJ_NULL_BRUSH = {
	OBJ_BRUSH, BS_NULL, RGB(0, 0, 0), TRUE
};

static struct hgdiobj OBJ_WHITE_PEN = {
	OBJ_PEN, PS_SOLID, RGB(255, 255, 255), TRUE
};

static struct hgdiobj OBJ_BLACK_PEN = {
	OBJ_PEN, PS_SOLID, RGB(0, 0, 0), TRUE
};

static struct hgdiobj OBJ_NULL_PEN = {
	OBJ_PEN, PS_NULL, RGB(0, 0, 0), TRUE
};

/* compiled fonts, numbered for now*/
#define FONT_SYSTEM_VAR		0	/* winSystem*/
#define FONT_GUI_VAR		1	/* winMSSansSerif*/
#define FONT_OEM_FIXED		2	/* fix: rom8x16*/
#define FONT_SYSTEM_FIXED	0	/* fix: winSystem*/
#define FONT_DEVICE_DEFAULT	2	/* rom8x16*/

static struct hgdiobj OBJ_OEM_FIXED_FONT = {
	OBJ_FONT, FONT_OEM_FIXED, 0L, TRUE
};

static struct hgdiobj OBJ_ANSI_FIXED_FONT = {
	OBJ_FONT, FONT_SYSTEM_FIXED, 0L, TRUE
};

static struct hgdiobj OBJ_ANSI_VAR_FONT = {
	OBJ_FONT, FONT_SYSTEM_VAR, 0L, TRUE
};

static struct hgdiobj OBJ_SYSTEM_FONT = {
	OBJ_FONT, FONT_SYSTEM_VAR, 0L, TRUE
};

static struct hgdiobj OBJ_DEVICE_DEFAULT_FONT = {
	OBJ_FONT, FONT_DEVICE_DEFAULT, 0L, TRUE
};

static struct hgdiobj OBJ_SYSTEM_FIXED_FONT = {
	OBJ_FONT, FONT_SYSTEM_FIXED, 0L, TRUE
};

static struct hgdiobj OBJ_DEFAULT_GUI_FONT = {
	OBJ_FONT, FONT_GUI_VAR, 0L, TRUE
};

static struct hgdiobj *stockObjects[MAXSTOCKOBJECTS] = {
	&OBJ_WHITE_BRUSH,		/* WHITE_BRUSH*/
	&OBJ_LTGRAY_BRUSH,		/* LTGRAY_BRUSH*/
	&OBJ_GRAY_BRUSH,		/* GRAY_BRUSH*/
	&OBJ_DKGRAY_BRUSH,		/* DKGRAY_BRUSH*/
	&OBJ_BLACK_BRUSH,		/* BLACK_BRUSH*/
	&OBJ_NULL_BRUSH,		/* NULL_BRUSH*/
	&OBJ_WHITE_PEN,			/* WHITE_PEN*/
	&OBJ_BLACK_PEN,			/* BLACK_PEN*/
	&OBJ_NULL_PEN,			/* NULL_PEN*/
	NULL,
	&OBJ_OEM_FIXED_FONT,		/* OEM_FIXED_FONT*/
	&OBJ_ANSI_FIXED_FONT,		/* ANSI_FIXED_FONT*/
	&OBJ_ANSI_VAR_FONT,		/* ANSI_VAR_FONT*/
	&OBJ_SYSTEM_FONT,		/* SYSTEM_FONT*/
	&OBJ_DEVICE_DEFAULT_FONT,	/* DEVICE_DEFAULT_FONT*/
	NULL,				/* DEFAULT_PALETTE*/
	&OBJ_SYSTEM_FIXED_FONT,		/* SYSTEM_FIXED_FONT*/
	&OBJ_DEFAULT_GUI_FONT		/* DEFAULT_GUI_FONT*/
};

HGDIOBJ WINAPI
GetStockObject(int nObject)
{
	if(nObject >= 0 && nObject < MAXSTOCKOBJECTS)
		return stockObjects[nObject];
	return NULL;
}

HGDIOBJ WINAPI
SelectObject(HDC hdc, HGDIOBJ hObject)
{
	HGDIOBJ	objOrg;

	if(!hdc || !hObject)
		return NULL;

	switch(hObject->type) {
	case OBJ_PEN:
		objOrg = hdc->pen;
		hdc->pen = hObject;
		break;
	case OBJ_BRUSH:
		objOrg = hdc->brush;
		hdc->brush = hObject;
		break;
	case OBJ_FONT:
		objOrg = hdc->font;
		hdc->font = hObject;
		break;
	default:
		return NULL;
	}

	return objOrg;
}

BOOL WINAPI
DeleteObject(HGDIOBJ hObject)
{
	if(!hObject || hObject->stockobj)
		return FALSE;

	ItemFree(hObject);
	return TRUE;
}

HBRUSH WINAPI
CreateSolidBrush(COLORREF crColor)
{
	HBRUSH	hbr;

	hbr = ItemNew(struct hgdiobj);
	hbr->type = OBJ_BRUSH;
	hbr->style = BS_SOLID;
	hbr->color = crColor;
	hbr->stockobj = FALSE;
	return hbr;
}

HPEN WINAPI
CreatePen(int nPenStyle, int nWidth, COLORREF crColor)
{
	HPEN	hpen;

	/* fix: nWidth > 1*/
	hpen = ItemNew(struct hgdiobj);
	hpen->type = OBJ_PEN;
	hpen->style = PS_SOLID;
	hpen->color = crColor;
	hpen->stockobj = FALSE;
	return hpen;
}
